"""Messages for application."""

from dataclasses import dataclass
from typing import Literal, Optional

from rich.console import RenderableType
from textual.message import Message
from textual.widget import Widget

from ..models.ollama_data import FullModel


@dataclass
class AppRequest(Message):
    """Request to app to perform an action."""

    widget: Optional[Widget]


@dataclass
class LocalModelCopied(Message):
    """Message to notify screen that local model has been copied."""

    src_model_name: str
    dst_model_name: str
    success: bool


@dataclass
class LocalModelCopyRequested(AppRequest):
    """Local model copy requested."""

    src_model_name: str
    dst_model_name: str


@dataclass
class SiteModelsLoaded(Message):
    """Message to notify screen that site models are loaded."""

    ollama_namespace: str


@dataclass
class SiteModelsRefreshRequested(AppRequest):
    """Site models refresh requested."""

    ollama_namespace: str
    force: bool


@dataclass
class StatusMessage(Message):
    """Message to update status bar."""

    msg: RenderableType
    log_it: bool = True


@dataclass
class PsMessage(Message):
    """Message to update ps status bar."""

    msg: RenderableType


@dataclass
class LocalModelListLoaded(Message):
    """Message to notify that local model list data is loaded."""


@dataclass
class ModelPulled(Message):
    """Message to notify screen that a model has been pulled."""

    model_name: str
    success: bool


@dataclass
class ModelPushed(Message):
    """Message to notify screen that a model has been pushed."""

    model_name: str
    success: bool


@dataclass
class ModelCreateRequested(AppRequest):
    """Message to notify that a model create has been requested."""

    model_name: str
    model_code: str
    quantization_level: Optional[str]


@dataclass
class ModelCreated(Message):
    """Message to notify that a model has been created."""

    model_name: str
    quantization_level: Optional[str]
    model_code: str
    success: bool


@dataclass
class ModelPullRequested(AppRequest):
    """Message to notify that a model pull has been requested."""

    model_name: str


@dataclass
class ModelPushRequested(AppRequest):
    """Message to notify that a model pull has been requested."""

    model_name: str


@dataclass
class LocalModelListRefreshRequested(AppRequest):
    """Message to notify that a local model list refresh has been requested."""


@dataclass
class LocalModelDeleted(Message):
    """Message to notify that a local model has been deleted."""

    model_name: str


@dataclass
class LocalModelDeleteRequested(AppRequest):
    """Message to notify that a local model delete has been requested."""

    model_name: str


@dataclass
class LocalModelDelete(Message):
    """Message to notify that a local model should be deleted."""

    model_name: str


@dataclass
class ShowLocalModel(Message):
    """Message to notify that a local model show has been requested."""

    model: FullModel


@dataclass
class NotifyInfoMessage(Message):
    """Message to toast info message."""

    message: str


@dataclass
class NotifyErrorMessage(Message):
    """Message to toast error message."""

    message: str


@dataclass
class SendToClipboard(Message):
    """Used to send a string to the clipboard."""

    message: str
    notify: bool = True


@dataclass
class SetModelNameLoading(Message):
    """Set model name loading."""

    model_name: str
    loading: bool


@dataclass
class ChangeTab(Message):
    """Change to requested tab."""

    tab: Literal["Local", "Site", "Tools", "Create", "Logs"]



"""Docker related functions"""

import os
from typing import List, Union

import docker.errors  # type: ignore
import docker.types  # type: ignore
from docker import DockerClient
from docker.models.containers import Container  # type: ignore

from .utils import read_env_file


# pylint: disable=too-many-arguments,too-many-return-statements,too-many-branches
def start_docker_container(
    image: str,
    *,
    container_name: str,
    command: str | None = None,
    env: dict | None = None,
    ports: dict | None = None,
    network_name: str | None = None,
    mounts: List[docker.types.Mount] | None = None,
    re_create: bool = False,
    remove: bool = False,
    background: bool = True,
) -> Union[str, Container, Exception]:
    """
    Start a docker container if it exists, otherwise create it

    Args:
        image (str): The base image to use
        container_name (str): The name of the container
        command (str | None): The command line to run. Defaults to None
        env (dict | None): The environment variables to use. Defaults to None
        ports (dict | None): The ports to expose. Defaults to None
        network_name (str | None): The network to use. Defaults to None
        mounts (List[docker.types.Mount]): The mounts to use.
            Example: [docker.types.Mount(target="/workspace", source=f"{os.getcwd()}/output/{session_id}", type="bind")]
        re_create (bool): Whether to recreate the container if it exists. Defaults to False
        remove (bool): Whether to remove the container after execution. Defaults to False
        background (bool): Whether to run the container in the background. Defaults to True

    Returns:
        bool: True if successful, False otherwise
    """

    try:
        if not env:
            env = {}
        env = read_env_file(".env") | env
        client: DockerClient
        if os.name == "nt":
            client = docker.DockerClient(base_url="tcp://127.0.0.1:2375", tls=False)
        else:
            client = docker.DockerClient(base_url="unix://var/run/docker.sock")
        if network_name:
            try:
                networks = client.networks.list(names=[network_name])
                if len(networks) > 0:
                    # network = networks[0]
                    print(f"Network {network_name} exists.")
                else:
                    client.networks.create(network_name, driver="bridge")
                    print(f"Network {network_name} created")
            except docker.errors.APIError as e:
                return e

        container: Container
        # container = client.containers.list(all=True, filters={"name": container_name})
        try:
            container = client.containers.get(container_name)
            # print(f"Container {container_name} exists.")
            if re_create:
                # print(f"Recreating container {container_name}")
                container.remove(force=True)
            else:
                if container.status == "running":
                    # print(f"Container {container_name} is already running.")
                    return container
                # print(f"Starting container {container_name}")
                container.start()
                return container
        except docker.errors.NotFound:
            pass
        except docker.errors.APIError as e:
            return e

        print(f"Container {container_name} does not exist. Creating...")
        if background:
            container = client.containers.run(
                image,
                command,
                name=container_name,
                ports=ports,
                detach=background,
                environment=env,
                remove=remove,
                network=network_name,
                mounts=mounts,
            )
            return container
        logs: bytes = client.containers.run(
            image,
            command,
            name=container_name,
            ports=ports,
            detach=background,
            environment=env,
            remove=remove,
            network=network_name,
            mounts=mounts,
        )
        return logs.decode("utf-8").strip()
    except docker.errors.DockerException as e:
        return e



"""Data manager for Par Llama."""

import os.path
import re
from datetime import datetime
from os import PathLike
from typing import (
    Any,
    Dict,
    Generator,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Union,
)

import docker  # type: ignore
import docker.errors  # type: ignore
import docker.types  # type: ignore
import ollama
import requests
import simplejson as json
from bs4 import BeautifulSoup
from docker.models.containers import Container  # type: ignore

from parllama.docker_utils import start_docker_container
from parllama.models.ollama_data import (
    FullModel,
    ModelListPayload,
    ModelShowPayload,
    SiteModel,
    SiteModelData,
)
from parllama.models.settings_data import settings
from parllama.utils import output_to_dicts, run_cmd
from parllama.widgets.local_model_list_item import LocalModelListItem
from parllama.widgets.site_model_list_item import SiteModelListItem


class DataManager:
    """Data manager for Par Llama."""

    ollama_site_categories: List[str] = ["popular", "featured", "newest"]
    models: List[LocalModelListItem]
    site_models: List[SiteModelListItem]

    def __init__(self):
        """ "Initialize the data manager."""
        self.models = []
        self.site_models = []

    @staticmethod
    def _get_all_model_data() -> List[LocalModelListItem]:
        """Get all model data."""
        all_models: List[LocalModelListItem] = []
        res = ModelListPayload(**ollama.list())
        pattern = r"^(# Modelfile .*)\n(# To build.*)\n# (FROM .*\n)\n(FROM .*)\n(.*)$"
        replacement = r"\3\5"
        for model in res.models:
            res2 = ModelShowPayload(**ollama.show(model.name))

            res2.modelfile = re.sub(
                pattern, replacement, res2.modelfile, flags=re.MULTILINE | re.IGNORECASE
            )
            res3 = FullModel(**model.model_dump(), **res2.model_dump())
            all_models.append(LocalModelListItem(res3))
        return all_models

    def refresh_models(self) -> List[LocalModelListItem]:
        """Refresh all local model data."""
        self.models = self._get_all_model_data()
        return self.models

    @staticmethod
    def model_ps() -> List[dict[str, Any]]:
        """Get model ps."""
        ret = run_cmd(["ollama", "ps"])

        if not ret:
            return []
        return output_to_dicts(ret)

    @staticmethod
    def pull_model(model_name: str) -> Iterator[Dict[str, Any]]:
        """Pull a model."""
        return ollama.pull(model_name, stream=True)  # type: ignore

    @staticmethod
    def push_model(model_name: str) -> Iterator[Dict[str, Any]]:
        """Push a model."""
        return ollama.push(model_name, stream=True)  # type: ignore

    def delete_model(self, model_name: str) -> bool:
        """Delete a model."""
        ret = ollama.delete(model_name).get("status", False)
        # ret = True
        if not ret:
            return False

        for model in self.models:
            if model.model.name == model_name:
                # model.remove()
                self.models.remove(model)
                return True

        return False

    def list_cache_files(self) -> List[str]:
        """List cache files."""
        if not os.path.exists(settings.cache_dir):
            return []

        return [
            f.split("-")[1].split(".")[0]
            for f in os.listdir(settings.cache_dir)
            if os.path.isfile(os.path.join(settings.cache_dir, f))
            and f.lower().endswith(".json")
        ]

    # pylint: disable=too-many-branches
    def refresh_site_models(
        self,
        namespace: str,
        category: Optional[Literal["popular", "featured", "newest"]] = None,
        force: bool = True,
    ) -> List[SiteModelListItem]:
        """Get list of all available models from Ollama.com."""

        settings.ensure_cache_folder()

        if not namespace:
            namespace = "models"
        namespace = os.path.basename(namespace)

        file_name = os.path.join(settings.cache_dir, f"site_models-{namespace}.json")
        if not force and os.path.exists(file_name):
            try:
                with open(file_name, "r", encoding="utf-8") as f:
                    ret: SiteModelData = SiteModelData(**json.load(f))
                    if ret.last_update and ret.last_update.timestamp() > (
                        ret.last_update.timestamp() - 60 * 60 * 24
                    ):
                        self.site_models = [SiteModelListItem(m) for m in ret.models]
                        return self.site_models
            except Exception as e:  # pylint: disable=broad-exception-caught
                print(f"Error: {e}")

        url_base: str = f"https://ollama.com/{namespace}"
        models: List[SiteModel] = []

        for cat in self.ollama_site_categories:
            if category and category != cat:
                continue
            url = url_base
            if namespace == "models":
                url += f"?sort={cat}"
            response = requests.get(
                url, headers={"User-Agent": "Mozilla/5.0"}, timeout=10
            )
            soup = BeautifulSoup(response.content.decode("utf-8"), "html.parser")
            for card in soup.find_all("li", class_="items-baseline"):
                meta_data = {
                    "name": card.find("h2").text.strip(),
                    "description": card.find("p").text.strip(),
                    "model_url": card.find("a")["href"],
                    "url": f'{url}{card.find("a")["href"]}',
                    "num_pulls": "",
                    "num_tags": "",
                    "updated": "",
                }

                pres = card.find_all(
                    "span", class_=["flex", "items-center"], recursive=True
                )
                pres = [p.text.strip() for p in pres]
                pres = [p for p in pres if p]

                for p in pres:
                    if "Pulls" in p:
                        meta_data["num_pulls"] = p.split("\n")[0].strip()
                    if "Tags" in p:
                        meta_data["num_tags"] = p.split("\xa0")[0].strip()
                    if "Updated" in p:
                        meta_data["updated"] = p.split("\xa0")[-1].strip()
                # tags = []
                tags = card.find_all("span", class_=["text-blue-600"], recursive=True)
                tags = [p.text.strip() for p in tags]
                tags = [p for p in tags if p]
                meta_data["tags"] = tags

                model = SiteModel(**meta_data)
                # print(model.model_dump_json(indent=4))
                found = False
                for m in models:
                    if m.name == model.name:
                        found = True
                        break
                if not found:
                    models.append(model)

            if namespace != "models":
                break

        if len(models) > 0 and not settings.no_save:
            with open(file_name, "w", encoding="utf-8") as f:
                f.write(
                    SiteModelData(
                        models=models, last_update=datetime.now()
                    ).model_dump_json(indent=4)
                )
        self.site_models = [SiteModelListItem(m) for m in models]
        return self.site_models

    @staticmethod
    def create_model(
        model_name: str,
        model_code: str | PathLike,
        quantize_level: Optional[str] = None,
    ) -> Iterator[Dict[str, Any]]:
        """Create a new model."""
        return ollama.create(
            model=model_name, modelfile=str(model_code), quantize=quantize_level, stream=True
        )  # type: ignore

    @staticmethod
    def copy_model(src_name: str, dst_name: str) -> Mapping[str, Any]:
        """Copy local model to new name"""
        return ollama.copy(source=src_name, destination=dst_name)

    @staticmethod
    def quantize_model(
        model_name: str, quantize_level: str = "q4_0"
    ) -> Union[str, Container, Generator[bytes, None, None]]:
        """
        Quantize a model

        Args:
            model_name (str): The name of the model to quantize
            quantize_level (str): The quantization level to use. Defaults to "q4_0"

        Returns:
            bool: True if successful, False otherwise
        """
        model_name = os.path.basename(model_name)
        model_folder = os.path.join(settings.data_dir, "quantize_workspace", model_name)
        if not os.path.exists(model_folder):
            raise FileNotFoundError(f"Model folder does not exist: {model_folder}")
        quantized_model_file = os.path.join(model_folder, f"{quantize_level}.bin")
        if os.path.exists(quantized_model_file):
            os.unlink(quantized_model_file)

        # docker run --rm -v .:/model ollama/quantize -q q4_0 /model
        ret = start_docker_container(
            image="ollama/quantize",
            container_name="parllamaQtz",
            command=f"-q {quantize_level} /model",
            network_name="parllama-network",
            mounts=[
                docker.types.Mount(
                    target="/model",
                    source=model_folder,
                    type="bind",
                )
            ],
            background=False,
        )
        if isinstance(ret, Exception):
            raise ret

        if not os.path.exists(quantized_model_file):
            raise FileNotFoundError(
                f"Quantized model does not exist: {quantized_model_file}"
            )
        if isinstance(ret, str):
            return ret

        return ret.logs(stream=True)


dm: DataManager = DataManager()



"""Create new model view."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, VerticalScroll
from textual.widgets import Button, Input, Label, TextArea

from parllama.dialogs.error_dialog import ErrorDialog
from parllama.messages.main import ModelCreateRequested


class ModelCreateView(Container):
    """Create new model view."""

    DEFAULT_CSS = """
    ModelCreateView {
      #name_quantize_row {
        height: 3;
        #model_name{
          width: 2fr;
          height: 3;
        }
        #ql {
          height: 3;
          width: 1fr;
          Label {
            width: 16;
            height: 1;
            margin: 1;
          }
          #quantize_level {
            width: 1fr;
            height: 3;
          }
        }
      }
      #editor {
        border: double $background;
        border-title-color: $accent;
      }
      #editor:focus {
        border: double $accent;
      }
    }
	"""

    BINDINGS = []

    text_area: TextArea

    def __init__(self, **kwargs) -> None:
        """
        Initialise the screen.
        """
        super().__init__(**kwargs)
        self.sub_title = "Create Model"
        self.name_input = Input(id="model_name", placeholder="Model Name")
        self.quantize_input = Input(
            id="quantize_level",
            placeholder="e.g. q4_0 or blank for none",
        )
        self.text_area = TextArea.code_editor("", id="editor")
        self.text_area.indent_type = "tabs"
        self.text_area.border_title = "Model Code"
        self.create_button = Button("Create", id="create_button")

    def compose(self) -> ComposeResult:
        """Compose the content of the screen."""
        with VerticalScroll(id="main_scroll"):
            with Horizontal(id="name_quantize_row"):
                yield self.name_input
                with Horizontal(id="ql"):
                    yield Label("Quantize Level")
                    yield self.quantize_input
            yield self.text_area
            yield self.create_button

    def on_mount(self) -> None:
        """Configure the dialog once the DOM is ready."""
        # self.name_input.focus()

    @on(Button.Pressed, "#create_button")
    def action_create_model(self) -> None:
        """Create the model."""
        name = (self.name_input.value or "").strip()
        code = (self.text_area.text or "").strip()
        quantization_level = (self.quantize_input.value or "").strip()
        if not name:
            self.app.push_screen(
                ErrorDialog(title="Input Error", message="Please enter a model name")
            )
            return
        if not code:
            self.app.push_screen(
                ErrorDialog(title="Input Error", message="Please enter a model code")
            )
            return
        self.post_message(
            ModelCreateRequested(
                widget=self,
                model_name=name,
                model_code=code,
                quantization_level=quantization_level,
            )
        )



"""Functions for formatting timestamps and converting datetimes to the local timezone."""

from datetime import datetime, timezone, tzinfo

import simplejson as json


def format_datetime(dt: datetime | None, fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
    """Convert a datetime object into a string in human-readable format.

    Args:
            dt: The datetime object.
            fmt: defaults to "%Y-%m-%d %H:%M:%S"

    Returns:
            The string datetime in the format specified. Or never if dt is None.
    """
    if dt is None:
        return "Never"
    return dt.strftime(fmt)


def format_timestamp(timestamp: float, fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
    """Convert a Unix timestamp into a string in human-readable format.

    Args:
            timestamp: The Unix timestamp.
            fmt: defaults to "%Y-%m-%d %H:%M:%S"

    Returns:
            The string timestamp in the format specified.
    """
    utc_dt = datetime.fromtimestamp(timestamp, timezone.utc)
    local_dt = utc_dt.astimezone()
    return local_dt.strftime(fmt)


def convert_to_local(utc_dt: datetime | str | None) -> datetime | None:
    """Given a UTC datetime, return a datetime in the local timezone."""
    if utc_dt is None:
        return None
    if isinstance(utc_dt, str):
        if utc_dt == "":
            return None
        utc_dt = datetime.fromisoformat(utc_dt)

    local_dt_now = datetime.now()
    local_tz = local_dt_now.astimezone().tzinfo
    local_dt = utc_dt.astimezone(local_tz)
    return local_dt


def get_local_timezone() -> tzinfo | None:
    """Return the local timezone."""
    return datetime.now(timezone.utc).astimezone().tzinfo


class DateTimeEncoder(json.JSONEncoder):
    """JSONEncoder subclass that knows how to encode datetime objects."""

    def default(self, o):  # type: ignore
        if isinstance(o, datetime):
            return o.isoformat()
        return super().default(o)



"""The main entry point for the application."""

from __future__ import annotations

import os

from parllama.app import ParLlamaApp
from parllama.models.settings_data import settings

if os.environ.get("DEBUG"):
    import pydevd_pycharm  # type: ignore

    pydevd_pycharm.settrace(
        "localhost", port=12345, suspend=False, patch_multiprocessing=True
    )


def run() -> None:
    """Run the application."""
    print(f"Settings folder {settings.data_dir}")
    ParLlamaApp().run()


if __name__ == "__main__":
    run()



"""The main application class."""

import asyncio
from queue import Empty, Queue
from typing import Any, Dict, Iterator, List, Set

import ollama
import pyperclip  # type: ignore
from rich.columns import Columns
from rich.console import RenderableType
from rich.progress_bar import ProgressBar
from rich.style import Style
from rich.text import Text
from textual import on, work
from textual.app import App
from textual.binding import Binding
from textual.color import Color
from textual.message import Message
from textual.notifications import SeverityLevel
from textual.widget import Widget
from textual.widgets import Input, Select, TextArea

from parllama import __application_title__
from parllama.data_manager import dm
from parllama.dialogs.help_dialog import HelpDialog
from parllama.messages.main import (
    AppRequest,
    ChangeTab,
    LocalModelCopied,
    LocalModelCopyRequested,
    LocalModelDelete,
    LocalModelDeleted,
    LocalModelListLoaded,
    LocalModelListRefreshRequested,
    ModelCreated,
    ModelCreateRequested,
    ModelPulled,
    ModelPullRequested,
    ModelPushed,
    ModelPushRequested,
    NotifyErrorMessage,
    NotifyInfoMessage,
    PsMessage,
    SendToClipboard,
    SetModelNameLoading,
    SiteModelsLoaded,
    SiteModelsRefreshRequested,
    StatusMessage,
)
from parllama.models.jobs import (
    CopyModelJob,
    CreateModelJob,
    PullModelJob,
    PushModelJob,
    QueueJob,
)
from parllama.models.settings_data import settings
from parllama.screens.main_screen import MainScreen
from parllama.theme_manager import theme_manager


class ParLlamaApp(App[None]):
    """Main application class"""

    TITLE = __application_title__
    BINDINGS = [
        Binding(key="f1", action="help", description="Help", show=True, priority=True),
        Binding(key="ctrl+q", action="app.quit", description="Quit", show=True),
        Binding(
            key="f10",
            action="toggle_dark",
            description="Toggle Dark Mode",
            show=True,
            priority=True,
        ),
        Binding(key="ctrl+c", action="noop", show=False),
    ]

    commands: list[dict[str, str]] = [
        {
            "action": "action_quit",
            "cmd": "Quit Application",
            "help": "Quit the application as soon as possible",
        }
    ]
    CSS_PATH = "app.tcss"

    DEFAULT_CSS = """
    """

    notify_subs: Dict[str, Set[Widget]]
    main_screen: MainScreen
    job_queue: Queue[QueueJob]
    is_busy: bool = False
    last_status: RenderableType = ""

    def __init__(self) -> None:
        """Initialize the application."""
        super().__init__()
        self.notify_subs = {"*": set[Widget]()}

        self.title = __application_title__
        self.dark = settings.theme_mode != "light"
        self.design = theme_manager.get_theme(settings.theme_name)  # type: ignore
        self.job_queue = Queue[QueueJob]()
        self.is_busy = False
        self.is_refreshing = False
        self.last_status = ""
        self.main_screen = MainScreen()

    def _watch_dark(self, value: bool) -> None:
        """Watch the dark property."""
        settings.theme_mode = "dark" if value else "light"
        settings.save_settings_to_file()

    def get_css_variables(self) -> dict[str, str]:
        """Get a mapping of variables used to pre-populate CSS.

        May be implemented in a subclass to add new CSS variables.

        Returns:
            A mapping of variable name to value.
        """
        return theme_manager.get_color_system_for_theme_mode(
            settings.theme_name, self.dark
        ).generate()

    @on(NotifyInfoMessage)
    def notify_info(self, msg: NotifyInfoMessage) -> None:
        """Show info toast message for 3 seconds"""
        self.notify(msg.message, timeout=3)

    @on(NotifyErrorMessage)
    def notify_error(self, msg: NotifyErrorMessage) -> None:
        """Show error toast message for 6 seconds"""
        self.notify(msg.message, severity="error", timeout=6)

    async def on_mount(self) -> None:
        """Display the main or locked screen."""
        await self.push_screen(self.main_screen)

        self.set_timer(1, self.do_jobs)
        self.set_timer(1, self.update_ps)

    def action_noop(self) -> None:
        """Do nothing"""

    def action_help(self) -> None:
        """Show help screen"""
        self.app.push_screen(HelpDialog())

    def action_clear_field(self) -> None:
        """Clear focused widget value"""
        f: Widget | None = self.screen.focused
        if not f:
            return
        if isinstance(f, Input):
            f.value = ""

        if isinstance(f, TextArea):
            f.text = ""

    def action_copy_to_clipboard(self) -> None:
        """Copy focused widget value to clipboard"""
        f: Widget | None = self.screen.focused
        if not f:
            return

        if isinstance(f, (Input, Select)):
            self.app.post_message(SendToClipboard(str(f.value) if f.value else ""))

        if isinstance(f, TextArea):
            self.app.post_message(SendToClipboard(f.selected_text or f.text))

    def action_cut_to_clipboard(self) -> None:
        """Cut focused widget value to clipboard"""
        f: Widget | None = self.screen.focused
        if not f:
            return
        if isinstance(f, Input):
            pyperclip.copy(f.value)
            f.value = ""
        if isinstance(f, TextArea):
            pyperclip.copy(f.selected_text or f.text)
            f.text = ""

    @on(SendToClipboard)
    def send_to_clipboard(self, msg: SendToClipboard) -> None:
        """Send string to clipboard"""
        # works for remote ssh sessions
        self.copy_to_clipboard(msg.message)
        # works for local sessions
        pyperclip.copy(msg.message)
        if msg.notify:
            self.post_message(NotifyInfoMessage("Value copied to clipboard"))

    @on(ModelPushRequested)
    def on_model_push_requested(self, msg: ModelPushRequested) -> None:
        """Push requested model event"""
        self.job_queue.put(PushModelJob(modelName=msg.model_name))
        self.post_message_all(SetModelNameLoading(msg.model_name, True))
        # self.notify(f"Model push {msg.model_name} requested")

        # primary_style = Style(
        #     color=theme_manager.get_color_system_for_theme_mode(
        #         self.theme, self.dark
        #     ).primary.rich_color
        # )
        # background_style = Style(
        #     color=(
        #             theme_manager.get_color_system_for_theme_mode(
        #                 self.theme, self.dark
        #             ).surface

    @on(ModelCreateRequested)
    def on_model_create_requested(self, msg: ModelCreateRequested) -> None:
        """Create model requested event"""
        self.job_queue.put(
            CreateModelJob(
                modelName=msg.model_name,
                modelCode=msg.model_code,
                quantizationLevel=msg.quantization_level,
            )
        )

    @on(LocalModelDelete)
    def on_local_model_delete(self, msg: LocalModelDelete) -> None:
        """Delete model event"""
        if not dm.delete_model(msg.model_name):
            self.post_message_all(SetModelNameLoading(msg.model_name, False))
            self.status_notify(
                f"Error deleting model {msg.model_name}.", severity="error"
            )
            return
        self.post_message_all(LocalModelDeleted(msg.model_name))

    @on(LocalModelDeleted)
    def on_model_deleted(self, msg: LocalModelDeleted) -> None:
        """Model deleted event"""
        self.status_notify(f"Model {msg.model_name} deleted.")

    @on(ModelPullRequested)
    def on_model_pull_requested(self, msg: ModelPullRequested) -> None:
        """Pull requested model event"""
        self.job_queue.put(PullModelJob(modelName=msg.model_name))
        self.post_message_all(SetModelNameLoading(msg.model_name, True))
        # self.notify(f"Model pull {msg.model_name} requested")

        # primary_style = Style(
        #     color=theme_manager.get_color_system_for_theme_mode(
        #         self.theme, self.dark
        #     ).primary.rich_color
        # )
        # background_style = Style(
        #     color=(
        #             theme_manager.get_color_system_for_theme_mode(
        #                 self.theme, self.dark
        #             ).surface
        #             or Color.parse("#111")
        #     ).rich_color
        # )
        # self.screen.post_message(
        #     StatusMessage(
        #         Columns(
        #             [
        #                 Text.assemble("status", " "),
        #                 ProgressBar(
        #                     total=100,
        #                     completed=50,
        #                     width=40,
        #                     style=background_style,
        #                     complete_style=primary_style,
        #                     finished_style=primary_style,
        #                 ),
        #             ]
        #         )
        #     )
        # )

    @on(LocalModelCopyRequested)
    def on_local_model_copy_requested(self, msg: LocalModelCopyRequested) -> None:
        """Local model copy request event"""
        self.job_queue.put(
            CopyModelJob(modelName=msg.src_model_name, dstModelName=msg.dst_model_name)
        )

    async def do_copy_local_model(self, msg: CopyModelJob) -> None:
        """Copy local model"""
        ret = dm.copy_model(msg.modelName, msg.dstModelName)
        self.post_message_all(
            LocalModelCopied(
                src_model_name=msg.modelName,
                dst_model_name=msg.dstModelName,
                success=ret["status"] == "success",
            )
        )

    @on(LocalModelCopied)
    def on_local_model_copied(self, msg: LocalModelCopied) -> None:
        """Local model copied event"""
        if msg.success:
            self.status_notify(
                f"Model {msg.src_model_name} copied to {msg.dst_model_name}"
            )
        else:
            self.status_notify(
                f"Copying model {msg.src_model_name} to {msg.dst_model_name} failed",
                severity="error",
            )

    async def do_progress(self, job: QueueJob, res: Iterator[Dict[str, Any]]) -> str:
        """Update progress bar"""
        try:
            last_status = ""
            for msg in res:
                last_status = msg["status"]
                pb: ProgressBar | None = None
                if "total" in msg and "completed" in msg:
                    msg["percent"] = (
                        str(int(msg["completed"] / msg["total"] * 100)) + "%"
                    )
                    primary_style = Style(
                        color=theme_manager.get_color_system_for_theme_mode(
                            settings.theme_name, self.dark
                        ).primary.rich_color
                    )
                    background_style = Style(
                        color=(
                            theme_manager.get_color_system_for_theme_mode(
                                settings.theme_name, self.dark
                            ).surface
                            or Color.parse("#111")
                        ).rich_color
                    )
                    pb = ProgressBar(
                        total=msg["total"],
                        completed=msg["completed"],
                        width=25,
                        style=background_style,
                        complete_style=primary_style,
                        finished_style=primary_style,
                    )
                else:
                    msg["percent"] = ""
                if msg["percent"] and msg["status"] == "success":
                    msg["percent"] = "100%"
                parts: List[RenderableType] = [
                    Text.assemble(
                        job.modelName,
                        " ",
                        msg["status"],
                        " ",
                        msg["percent"],
                        " ",
                    )
                ]
                if pb:
                    parts.append(pb)

                self.main_screen.post_message(
                    StatusMessage(Columns(parts), log_it=False)
                )
            return last_status
        except ollama.ResponseError as e:
            self.main_screen.post_message(
                StatusMessage(Text.assemble(("error:" + str(e), "red")))
            )
            raise e

    async def do_pull(self, job: PullModelJob) -> None:
        """Pull a model"""
        try:
            res = dm.pull_model(job.modelName)
            last_status = await self.do_progress(job, res)

            self.post_message_all(
                ModelPulled(model_name=job.modelName, success=last_status == "success")
            )
        except ollama.ResponseError:
            self.post_message_all(ModelPulled(model_name=job.modelName, success=False))

    async def do_push(self, job: PushModelJob) -> None:
        """Push a model"""
        try:
            res = dm.push_model(job.modelName)
            last_status = await self.do_progress(job, res)

            self.post_message_all(
                ModelPushed(model_name=job.modelName, success=last_status == "success")
            )
        except ollama.ResponseError:
            self.post_message_all(ModelPushed(model_name=job.modelName, success=False))

    async def do_create_model(self, job: CreateModelJob) -> None:
        """Create a model"""
        try:
            res = dm.create_model(job.modelName, job.modelCode, job.quantizationLevel)
            last_status = await self.do_progress(job, res)

            self.post_message_all(
                ModelCreated(
                    model_name=job.modelName,
                    model_code=job.modelCode,
                    quantization_level=job.quantizationLevel,
                    success=last_status == "success",
                )
            )
        except ollama.ResponseError:
            self.post_message_all(
                ModelCreated(
                    model_name=job.modelName,
                    model_code=job.modelCode,
                    quantization_level=job.quantizationLevel,
                    success=False,
                )
            )

    @work(group="do_jobs", thread=True)
    async def do_jobs(self) -> None:
        """Pull all requested models in queue"""
        while True:
            try:
                job: QueueJob = self.job_queue.get(block=True, timeout=1)
                if self._exit:
                    return
                if not job:
                    continue
                self.is_busy = True
                if isinstance(job, PullModelJob):
                    await self.do_pull(job)
                elif isinstance(job, PushModelJob):
                    await self.do_push(job)
                elif isinstance(job, CopyModelJob):
                    await self.do_copy_local_model(job)
                elif isinstance(job, CreateModelJob):
                    await self.do_create_model(job)
                else:
                    self.status_notify(
                        f"Unknown job type {type(job)}",
                        severity="error",
                    )
            except Empty:
                if self._exit:
                    return
                if self.is_busy:
                    self.post_message(LocalModelListRefreshRequested(widget=None))
                    self.is_busy = False
                continue

    @on(ModelPulled)
    def on_model_pulled(self, msg: ModelPulled) -> None:
        """Model pulled event"""
        if msg.success:
            self.status_notify(
                f"Model {msg.model_name} pulled.",
            )
        else:
            self.status_notify(
                f"Model {msg.model_name} failed to pull.",
                severity="error",
            )

    @on(ModelCreated)
    def on_model_created(self, msg: ModelCreated) -> None:
        """Model created event"""
        if msg.success:
            self.status_notify(
                f"Model {msg.model_name} created.",
            )
            self.set_timer(1, self.action_refresh_models)
        else:
            self.status_notify(
                f"Model {msg.model_name} failed to create.",
                severity="error",
            )

    def action_refresh_models(self) -> None:
        """Refresh models action."""
        self.refresh_models()

    @on(AppRequest)
    def on_app_request(self, msg: AppRequest) -> None:
        """Add any widget that requests an action to notify_subs"""
        if msg.widget:
            self.notify_subs["*"].add(msg.widget)

    @on(LocalModelListRefreshRequested)
    def on_model_list_refresh_requested(self) -> None:
        """Model refresh request event"""
        if self.is_refreshing:
            self.status_notify("A model refresh is already in progress. Please wait.")
            return
        self.refresh_models()

    @work(group="refresh_models", thread=True)
    async def refresh_models(self):
        """Refresh the models."""
        self.is_refreshing = True
        try:
            self.main_screen.post_message(
                StatusMessage("Local model list refreshing...")
            )
            dm.refresh_models()
            self.main_screen.post_message(StatusMessage("Local model list refreshed"))
            self.post_message_all(LocalModelListLoaded())
        finally:
            self.is_refreshing = False

    @on(LocalModelListLoaded)
    def on_model_data_loaded(self) -> None:
        """Refresh model completed"""
        self.main_screen.post_message(StatusMessage("Local model list refreshed"))
        # self.notify("Local models refreshed.")

    @on(SiteModelsRefreshRequested)
    def on_site_models_refresh_requested(self, msg: SiteModelsRefreshRequested) -> None:
        """Site model refresh request event"""
        if self.is_refreshing:
            self.status_notify("A model refresh is already in progress. Please wait.")
            return
        self.refresh_site_models(msg)

    @on(SiteModelsLoaded)
    def on_site_models_loaded(self, msg: SiteModelsLoaded) -> None:
        """Site model refresh completed"""
        self.status_notify(
            f"Site models refreshed for {msg.ollama_namespace or 'models'}"
        )

    @work(group="refresh_site_model", thread=True)
    async def refresh_site_models(self, msg: SiteModelsRefreshRequested):
        """Refresh the site model."""
        self.is_refreshing = True
        try:
            self.main_screen.post_message(
                StatusMessage(
                    f"Site models for {msg.ollama_namespace or 'models'} refreshing... force={msg.force}"
                )
            )
            dm.refresh_site_models(msg.ollama_namespace, None, msg.force)
            self.post_message_all(
                SiteModelsLoaded(ollama_namespace=msg.ollama_namespace)
            )
            self.main_screen.post_message(
                StatusMessage(
                    f"Site models for {msg.ollama_namespace or 'models'} loaded. force={msg.force}"
                )
            )

        finally:
            self.is_refreshing = False

    @work(group="update_ps", thread=True)
    async def update_ps(self) -> None:
        """Update ps msg"""
        was_blank = False
        while self.is_running:
            await asyncio.sleep(1)
            ret = dm.model_ps()
            if not ret:
                if not was_blank:
                    self.main_screen.post_message(PsMessage(msg=""))
                was_blank = True
                continue
            was_blank = False
            info = ret[0]
            self.main_screen.post_message(
                PsMessage(
                    msg=Text.assemble(
                        "Name: ",
                        info["name"],
                        " Size: ",
                        info["size"],
                        " Processor: ",
                        info["processor"],
                        " Until: ",
                        info["until"],
                    )
                )
            )
        self.main_screen.post_message(StatusMessage(msg="exited..."))

    def status_notify(self, msg: str, severity: SeverityLevel = "information") -> None:
        """Show notification and update status bar"""
        self.notify(msg, severity=severity)
        self.main_screen.post_message(StatusMessage(msg))

    def post_message_all(self, msg: Message) -> None:
        """Post a message to all screens"""
        if isinstance(msg, StatusMessage):
            self.log(msg.msg)
            self.last_status = msg.msg
        for w in list(self.notify_subs["*"]):
            w.post_message(msg)
        if self.main_screen:
            self.main_screen.post_message(msg)

    @on(ChangeTab)
    def on_change_tab(self, msg: ChangeTab) -> None:
        """Change tab event"""
        msg.stop()
        self.main_screen.change_tab(msg.tab)



"""Local Model List Item."""

import humanize
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.widget import Widget
from textual.widgets import Static

from parllama.models.ollama_data import FullModel
from parllama.widgets.field_set import FieldSet


class LocalModelListItem(Widget):
    """Local Model List Item."""

    DEFAULT_CSS = """
    LocalModelListItem {
        padding: 1 2 1 2;
        width: 84;
        height: 1fr;
        background: $background;
        border: solid $accent;
        border-title-color: $primary;
        Static {
            background: transparent;
        }
        &.--highlight {
            border: double $primary;
            border-title-color: $primary-lighten-2;
            background: $panel;
            Static {
                background: transparent;
            }
        }

    }
    """
    model: FullModel

    def __init__(self, model: FullModel) -> None:
        """Initialize the item."""
        super().__init__()
        self.model = model
        self.can_focus = True

    def watch_has_focus(self, value: bool) -> None:
        """Watch the has_focus property and updated parent selected."""
        super().watch_has_focus(value)
        if value and hasattr(self.parent, "selected"):
            self.parent.selected = self if value else None  # type: ignore

    def compose(self) -> ComposeResult:
        """Compose the list item."""
        self.border_title = self.model.name
        with Vertical():
            # yield FieldSet("Name", Static(self.model.name, id="name"))
            yield FieldSet(
                "Modified", Static(str(self.model.modified_at), id="modified_at")
            )
            exp = str(self.model.expires_at)
            if exp in ["None", "0001-01-01 00:00:00+00:00"]:
                exp = "Never"
            yield FieldSet("Expires", Static(exp, id="expires_at"))
            yield FieldSet(
                "Size", Static(humanize.naturalsize(self.model.size), id="size")
            )
            yield Static("Digest:")
            yield Static(self.model.digest, id="digest")



"""Provides model details dialog."""

import re
from typing import List, cast

import humanize
import ollama
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import MarkdownViewer, Static, TextArea

from ..models.ollama_data import FullModel, MessageRoles
from ..widgets.field_set import FieldSet


class ModelDetailsDialog(ModalScreen[None]):
    """Modal dialog that shows model details."""

    DEFAULT_CSS = """
    ModelDetailsDialog {
        background: black 75%;
        align: center middle;
        &> VerticalScroll {
            background: $surface;
            width: 75%;
            height: 90%;
            min-width: 80;
            border: thick $accent;
            border-title-color: $primary;
            padding: 1;
        }
        .editor {
            background: $panel;
            border: solid $primary;
            border-title-color: $primary;
            margin-bottom: 1;
            .text-area--cursor-line {
              background: transparent;
            }
        }
    }
    """

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape, ctrl+q", "app.pop_screen", "", show=True),
        Binding("ctrl+c", "app.copy_to_clipboard", "", show=True),
    ]
    model: FullModel

    def __init__(self, model: FullModel) -> None:
        super().__init__()
        self.model = model

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        vs = VerticalScroll()
        vs.border_title = f"[ {self.model.name} ]"
        with vs:
            # yield FieldSet("Name", Static(self.model.name, id="name"))
            yield FieldSet(
                "Modified", Static(str(self.model.modified_at), id="modified_at")
            )
            exp = str(self.model.expires_at)
            if exp == "0001-01-01 00:00:00+00:00":
                exp = "Never"
            yield FieldSet("Expires", Static(exp, id="expires_at"))
            yield FieldSet(
                "Size", Static(humanize.naturalsize(self.model.size), id="size")
            )
            yield FieldSet("Digest", Static(self.model.digest, id="digest"))
            yield Static("")
            ta = TextArea(
                self.model.template or "", id="template", classes="editor height-auto"
            )
            ta.border_title = "Template"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.parameters or "",
                id="parameters",
                classes="editor height-auto",
            )
            ta.border_title = "Parameters"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.modelfile, id="modelfile", classes="editor height-10"
            )
            ta.border_title = "Model file"
            ta.read_only = True
            yield ta

            ta = TextArea(
                self.model.license or "", id="license", classes="editor height-10"
            )
            ta.border_title = "License"
            ta.read_only = True
            yield ta

            system_regex = re.compile(r"^system (.*)", re.I)
            message_regex = re.compile(r"^message (user|assistant|system) (.*)", re.I)
            messages: List[ollama.Message] = []
            system_msg: str = ""
            for line in self.model.modelfile.splitlines():
                match = message_regex.match(line)
                if match:
                    messages.append(
                        ollama.Message(
                            role=cast(MessageRoles, match.group(1)),
                            content=match.group(2),
                        )
                    )
                match = system_regex.match(line)
                if match:
                    system_msg = match.group(1)

            msgs = [f"* MESSAGE {m['role']} {m['content']}" for m in messages]
            if system_msg:
                msgs.insert(0, f"* SYSTEM {system_msg}")
            md = MarkdownViewer(
                "\n".join(msgs),
                id="messages",
                classes="editor height-auto",
                show_table_of_contents=False,
            )
            md.border_title = "Messages"
            yield md



"""Theme manager for Textual"""

import os
import shutil
from pathlib import Path
from typing import Dict, List, Literal, TypeAlias

import simplejson as json
from textual.design import ColorSystem

from parllama.models.settings_data import settings

ThemeMode: TypeAlias = Literal["dark", "light"]
ThemeModes: List[ThemeMode] = ["dark", "light"]
Theme: TypeAlias = Dict[ThemeMode, ColorSystem]
Themes: TypeAlias = Dict[str, Theme]


class InvalidThemeError(Exception):
    """Raised when an invalid theme is provided."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        self.theme_name = theme_name
        super().__init__(f"Invalid theme: {theme_name}")


class ThemeModeError(InvalidThemeError):
    """Raised when a theme does not have at least one of 'dark' or 'light' modes."""

    def __init__(self, theme_name: str):
        """Initialize the exception with the invalid theme name."""
        super().__init__(
            f"Theme '{theme_name}' does not have at least one of 'dark' or 'light' modes."
        )


class ThemeManager:
    """Theme manager for Textual"""

    themes: Themes
    theme_folder: str

    def __init__(self) -> None:
        """Initialize the theme manager"""
        self.theme_folder = os.path.join(settings.data_dir, "themes")

        self.themes = self.load_themes()

    def ensure_default_theme(self) -> None:
        """Ensure that the default theme exists."""
        default_theme_file: str = os.path.join(
            os.path.dirname(Path(__file__)), "themes", "par.json"
        )
        if not os.path.exists(self.theme_folder):
            os.makedirs(self.theme_folder)
        theme_file = os.path.join(self.theme_folder, "par.json")

        if not os.path.exists(theme_file):
            shutil.copy(default_theme_file, theme_file)

    def load_theme(self, theme_name: str) -> Theme:
        """Load textual theme from json file"""

        theme: Theme = {}
        theme_name = os.path.basename(theme_name)
        with open(
            os.path.join(self.theme_folder, theme_name), "r", encoding="utf-8"
        ) as f:
            theme_def = json.load(f)
            if "dark" not in theme_def and "light" not in theme_def:
                raise ThemeModeError(theme_name)

            for mode in ThemeModes:
                if mode in theme_def:
                    theme[mode] = ColorSystem(**theme_def[mode])
        return theme

    def load_themes(self) -> Themes:
        """Load textual themes from json files"""
        self.ensure_default_theme()

        themes: Themes = {}
        for file in os.listdir(self.theme_folder):
            if file.lower().endswith(".json"):
                theme_name = os.path.splitext(file)[0]
                themes[theme_name] = self.load_theme(f"{self.theme_folder}/{file}")
        return themes

    def get_theme(self, theme_name: str) -> Theme:
        """Get theme by name"""
        return self.themes[theme_name]

    def list_themes(self) -> List[str]:
        """Get list of themes"""
        return list(self.themes.keys())

    def theme_has_dark(self, theme_name: str) -> bool:
        """Check if theme has dark mode"""
        return "dark" in self.themes[theme_name]

    def theme_has_light(self, theme_name: str) -> bool:
        """Check if theme has light mode"""
        return "light" in self.themes[theme_name]

    def get_color_system_for_theme_mode(
        self, theme_name: str, dark: bool
    ) -> ColorSystem:
        """Get color system for theme mode"""
        theme = self.themes[theme_name]

        if dark:
            if "dark" in theme:
                return theme["dark"]
            return theme["light"]

        if "light" in theme:
            return theme["light"]
        return theme["dark"]


theme_manager = ThemeManager()



"""A label that can be clicked."""

from __future__ import annotations

from dataclasses import dataclass
from typing import cast

from rich.text import Text
from textual.message import Message
from textual.widgets import Label

from parllama.messages.main import SendToClipboard


class ClickableLabel(Label):
    """A label that can be clicked."""

    @dataclass
    class Click(Message):
        """A message that is sent when the label is clicked."""

        label: ClickableLabel
        """The label that was clicked."""

        @property
        def control(self) -> ClickableLabel:
            """Return the label that was clicked."""
            return self.label

    @property
    def plain_text(self) -> str:
        """Return the plain text of the label."""
        if isinstance(self.renderable, str):
            return self.renderable
        if isinstance(self.renderable, Text):
            return cast(Text, self.renderable).plain
        raise ValueError("Unknown renderable type")

    def on_click(self) -> None:
        """Called when the label is clicked."""
        self.post_message(self.Click(self))


class CopyToClipboardLabel(ClickableLabel):
    """A label that copies its text to the clipboard when clicked."""

    max_len: int
    """The maximum length of the text to copy to the clipboard. 0 = all"""

    def __init__(self, text: str | Text, max_len: int = 0, **kwargs) -> None:
        """Initialize the label."""
        super().__init__(text, **kwargs)
        self.max_len = max_len
        self.tooltip = "Copy to clipboard"

    @property
    def sanitized_text(self) -> str:
        """Return the text to be copied to the clipboard."""
        txt = self.plain_text
        if self.max_len > 0:
            txt = txt[: self.max_len]
        return txt

    def on_click(self) -> None:
        """Called when the label is clicked to post message to clipboard."""
        super().on_click()
        self.post_message(SendToClipboard(self.sanitized_text))



"""Widget for viewing application logs."""

from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import Button, Checkbox, Input, Label, RichLog

from parllama.models.settings_data import settings


class LogView(Container):
    """Widget for viewing application logs."""

    DEFAULT_CSS = """
    LogView {
      #tool_bar {
        height: 3;
        background: $surface-darken-1;
        #max_lines {
          width: 10;
        }
        Label {
          margin-top: 1;
          background: transparent;
        }
      }
      #logs {
        border: solid $primary;
      }
    }
    """

    richlog: RichLog
    max_lines_input: Input

    def __init__(self, **kwargs) -> None:
        """Initialise the screen."""
        super().__init__(**kwargs)
        self.sub_title = "Application Logs"
        self.richlog = RichLog(id="logs", wrap=True, highlight=True, auto_scroll=True)
        self.auto_scroll = Checkbox(label="Auto Scroll", value=True, id="auto_scroll")
        self.max_lines_input = Input(
            id="max_lines",
            placeholder="Max Lines",
            value=str(settings.max_log_lines),
            type="integer",
        )
        self.richlog.max_lines = int(self.max_lines_input.value)

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""

        with Vertical(id="menu"):
            with Horizontal(id="tool_bar"):
                yield self.auto_scroll
                yield Label("Max Lines: ")
                yield self.max_lines_input
                yield Button("Clear", id="clear", variant="warning")
            yield self.richlog

        self.richlog.write("Starting...")

    @on(Checkbox.Changed, "#auto_scroll")
    def on_auto_scroll_changed(self, event: Checkbox.Changed) -> None:
        """Handle auto scroll checkbox change"""
        self.richlog.auto_scroll = event.value

    @on(Button.Pressed, "#clear")
    def on_clear_pressed(self) -> None:
        """Handle clear button press"""
        self.richlog.clear()

    @on(Input.Changed, "#max_lines")
    def on_max_lines_changed(self, event: Input.Changed) -> None:
        """Handle max lines input change"""
        max_lines: int = max(0, int(event.value or "0"))
        self.richlog.max_lines = max_lines
        settings.max_log_lines = max_lines
        settings.save_settings_to_file()



"""Double clickable List Item."""

import time

from textual import events
from textual.message import Message
from textual.widgets import ListItem


class DblClickListItem(ListItem):
    """Double clickable List Item."""

    last_click: float = 0

    class DoubleClicked(Message):
        """fire when the item is double-clicked"""

        def __init__(self, item: ListItem) -> None:
            self.item = item
            super().__init__()

    async def _on_click(self, _: events.Click) -> None:
        """Handle the click event and check for double click."""
        if time.time() - self.last_click < 0.5:
            self.post_message(self.DoubleClicked(self))
            self.last_click = 0
            return
        await super()._on_click(_)
        self.last_click = time.time()



"""Main screen for TUI."""

from typing import Literal, cast

from rich.console import RenderableType
from textual import on
from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Footer, Header, Static, TabbedContent, TabPane

from parllama.messages.main import PsMessage, StatusMessage
from parllama.models.settings_data import ScreenType, settings
from parllama.widgets.local_model_view import LocalModelView
from parllama.widgets.log_view import LogView
from parllama.widgets.model_create_view import ModelCreateView
from parllama.widgets.model_tools_view import ModelToolsView
from parllama.widgets.site_model_view import SiteModelView


class MainScreen(Screen[None]):
    """Main screen."""

    BINDINGS = []

    CSS_PATH = "main_screen.tcss"

    status_bar: Static
    ps_status_bar: Static
    tabbed_content: TabbedContent
    log_view: LogView

    def __init__(self, **kwargs) -> None:
        """Initialize the Main screen."""
        super().__init__(**kwargs)
        self.status_bar = Static("", id="StatusBar")
        self.ps_status_bar = Static("", id="PsStatusBar")
        self.ps_status_bar.display = False
        self.log_view = LogView()

    async def on_mount(self) -> None:
        """Mount the Main screen."""
        self.change_tab(settings.starting_screen)
        self.set_timer(0.5, self.done_loading)

    def done_loading(self) -> None:
        """Hide loading indicator."""
        self.tabbed_content.loading = False

    def compose(self) -> ComposeResult:
        """Compose the Main screen."""
        yield Header(show_clock=True)
        yield Footer()
        yield self.status_bar
        yield self.ps_status_bar
        with TabbedContent(id="tabbed_content", initial=settings.starting_screen) as tc:
            self.tabbed_content = tc
            tc.loading = True

            with TabPane("Local", id="Local"):
                yield LocalModelView(id="local_models")
            with TabPane("Site", id="Site"):
                yield SiteModelView(id="site_models")
            with TabPane("Tools", id="Tools"):
                yield ModelToolsView(id="model_tools")
            with TabPane("Create", id="Create"):
                yield ModelCreateView(id="model_create")
            with TabPane("Logs", id="Logs"):
                yield self.log_view

    @on(TabbedContent.TabActivated)
    def on_tab_activated(self, msg: TabbedContent.TabActivated) -> None:
        """Tab activated event"""
        msg.stop()
        settings.last_screen = cast(ScreenType, msg.tab.label.plain)
        settings.save_settings_to_file()

    @on(StatusMessage)
    def on_status_message(self, msg: StatusMessage) -> None:
        """Status message event"""
        # msg.stop()
        self.update_status(msg.msg)
        if msg.log_it:
            self.log_view.richlog.write(msg.msg)

    def update_status(self, msg: RenderableType):
        """Update the status bar."""
        self.status_bar.update(msg)

    @on(PsMessage)
    def on_ps_message(self, msg: PsMessage) -> None:
        """PS status message event"""
        msg.stop()
        self.update_ps_status(msg.msg)

    def update_ps_status(self, msg: RenderableType):
        """Update the ps status bar."""
        self.ps_status_bar.update(msg)
        self.ps_status_bar.display = bool(msg)

    def change_tab(
        self, tab: Literal["Local", "Site", "Tools", "Create", "Logs"]
    ) -> None:
        """Change active tab."""
        # self.tabbed_content.active = tab

    def action_site_tag_clicked(self, model_tag: str) -> None:
        """Update search box with tag"""
        self.query_one("#site_models", SiteModelView).search_input.value = model_tag



"""Model for application settings."""

import os
import shutil
from argparse import Namespace
from typing import List, Literal, TypeAlias

import simplejson as json
from pydantic import BaseModel

from ..utils import get_args

ScreenType: TypeAlias = Literal["Local", "Site", "Tools", "Create", "Logs"]
valid_screens: List[ScreenType] = ["Local", "Site", "Tools", "Create", "Logs"]


class Settings(BaseModel):
    """Model for application settings."""

    no_save: bool = False
    data_dir: str = os.path.expanduser("~/.parllama")
    cache_dir: str = ""
    settings_file: str = "settings.json"
    theme_name: str = "par"
    starting_screen: ScreenType = "Local"
    last_screen: ScreenType = "Local"
    theme_mode: str = "dark"
    site_models_namespace: str = ""
    max_log_lines: int = 1000

    # pylint: disable=too-many-branches
    def __init__(self) -> None:
        """Initialize BwItemData."""
        super().__init__()
        args: Namespace = get_args()

        if args.no_save:
            self.no_save = True

        self.data_dir = (
            args.data_dir
            or os.environ.get("PARLLAMA_DATA_DIR")
            or os.path.expanduser("~/.parllama")
        )
        self.cache_dir = os.path.join(self.data_dir, "cache")
        os.makedirs(self.data_dir, exist_ok=True)

        if not os.path.exists(self.data_dir):
            raise FileNotFoundError(
                f"Par Llama data directory does not exist: {self.data_dir}"
            )

        self.settings_file = os.path.join(self.data_dir, "settings.json")
        if args.restore_defaults:
            if os.path.exists(self.settings_file):
                os.unlink(self.settings_file)
            theme_file = os.path.join(self.data_dir, "themes", "par.json")
            if os.path.exists(theme_file):
                os.unlink(theme_file)
        if args.clear_cache:
            if os.path.exists(self.cache_dir):
                shutil.rmtree(self.cache_dir, ignore_errors=True)

        self.load_from_file()
        if os.environ.get("PARLLAMA_THEME_NAME"):
            self.theme_name = os.environ.get("PARLLAMA_THEME_NAME", self.theme_name)

        if os.environ.get("PARLLAMA_THEME_MODE"):
            self.theme_mode = os.environ.get("PARLLAMA_THEME_MODE", self.theme_mode)

        if args.theme_name:
            self.theme_name = args.theme_name
        if args.theme_mode:
            self.theme_mode = args.theme_mode
        if args.starting_screen:
            self.starting_screen = args.starting_screen.capitalize()
            if self.starting_screen not in valid_screens:
                self.starting_screen = "Local"

        self.save_settings_to_file()

    def load_from_file(self) -> None:
        """Load settings from file."""
        try:
            with open(self.settings_file, mode="rt", encoding="utf-8") as f:
                data = json.load(f)
                self.theme_name = data.get("theme_name", self.theme_name)
                self.theme_mode = data.get("theme_mode", self.theme_mode)
                self.site_models_namespace = data.get("site_models_namespace", "")
                self.starting_screen = data.get("starting_screen", "Local")
                if self.starting_screen not in valid_screens:
                    self.starting_screen = "Local"

                self.last_screen = data.get("last_screen", "Local")
                if self.last_screen not in valid_screens:
                    self.last_screen = self.starting_screen

                self.max_log_lines = max(0, data.get("max_log_lines", 1000))
        except FileNotFoundError:
            pass  # If file does not exist, continue with default settings

    def save_settings_to_file(self) -> None:
        """Save settings to file."""
        if self.no_save:
            return
        os.makedirs(self.data_dir, exist_ok=True)
        if not os.path.exists(self.data_dir):
            raise FileNotFoundError(
                f"Par Llama data directory does not exist: {self.data_dir}"
            )

        with open(self.settings_file, "wt", encoding="utf-8") as f:
            f.write(self.model_dump_json(indent=4))

    def ensure_cache_folder(self) -> None:
        """Ensure the cache folder exists."""
        if not os.path.exists(self.cache_dir):
            os.mkdir(self.cache_dir)


settings = Settings()
